/*
 * dbtool - a simple but powerfull commandline interface
 *      to the GNU gdbm system, usefull for shellscripts
 *      which needs a database for data storage.
 *
 * Copyright (c) 2001 Thomas Linden, <tom@daemon.de>
 * All Rights Reserved.      Released under the GPL.
 *
 * $Id: config.cc,v 1.1 2002/02/11 14:33:26 scip Exp $
 */



#include "dbtool.h"
#include "config.h"



Config::Config() {
  /* konstructor */ 
  pkg = PACKAGE;
  default_db = DB;

  usage = 
"\nUsage: " + pkg + " -d database [DirusSfwVhtRFpP] [-k key] [-v value]
 -d database   Specify the filename of the database
 -i            Insert data, -k and -v is required (*)
 -r            Remove data, -k is required
 -u            Update data, -k and -v is required
 -s            Search for a key, -k is required
 -S            Search for all entrys which match a regular
               expression, the regexp must be specified with
               the -k parameter, i.e.: " + pkg + " -d file -S -k \"^[a-z]+\\d{3}.*$\"
 -D            Dump all keys/values of the specified gdbm file
 -w            Print out keys too, only possible with -S and -s
 -f            Force mode (overwrite file and data!)
 -k key        the key for an entry
 -v value      the value for the corresponding key, -k is required
 -F separator  field separator, optional used together with
               -i if no -k and -v is specified, and with -D as
               output separator
 -t expression Use expression to decide which part of an input line
               is to be used as the key and which one as the value, i.e.:
               " + pkg + " -i -d file -t \"^(.*)(\\/.*)$\";
               only possible with -i and without -k and -v(read from stdin)
 -R            Reverse meaning of token -t, the first match will be the
               value and the second one the key, if used with -s, -S or -D then
               first the value and then the key will be printed
 -p            Use encrypted database. " + pkg + " will ask you for the passphrase,
               unless the environment variable DB_PASSPHRASE is set.
 -P passphrase Use encrypted database. Specify the passphrase on the commandline
 -V            Print the version of dbtool
 -h            Print this usage message

If you insert data and omit -k and -v then " + pkg + " will read
the key/value pair from STDIN, key and value will be
separated by the parameter of the -F option or, if -F
is omited, by whitespace, the record separator is the
newline, thus one line stands for one key/value pair.

If no database filename is specified (-d) but the
environment variable DB_DATABASE exists, " + pkg + " will
use the file specified in DB_DATABASE.

" + pkg + " is free software under the terms of the GPL.
Report bugs to Thomas Linden <tom@daemon.de>";
}


void Config::args(int argc, char *argv[]) {
  _argc = argc;
  _argv = argv;
}

int Config::parse() {
  /* parse commandline parameters */

  int option;
  int cmd_insert = 0;

  separator = ' ';               /* default field separator, used by -S and -D and -i */
  command   = 'X';               /* this should get something else */
  with      = 0;                 /* do not print keys, used by -S and -s */
  reverse   = 0;                 /* first key, then value (not reverse is the default) */

  char *fname;
  fname = getenv(default_db);
  if(fname != NULL) {
    filename = fname;            /* will be overwritten by -d */
  }

  while (1) {
    option = getopt(_argc, _argv, "F:DSRft:wuirsd:k:v:VhpP:");
    if(option == -1) break;

    switch(option) {
    case 'f':
      force = 1;
      break;
    case 'i':
      command = 'i';
      break;
    case 'r':
      command = 'r';
      break;
    case 's':
      command = 's';
      readonly = 1;
      break;
    case 'S':
      command = 'S';
      readonly = 1;
      break;
    case 'u':
      command = 'u';
      break;
    case 'd':
      filename = optarg;
      break;
    case 'k':
      key = optarg;
      cmd_insert = 1;
      break;
    case 'v':
      value = optarg;
      cmd_insert = 1;
      break;
    case 'F':
      separator = optarg[0];
      break;
    case 'R':
      reverse = 1;
      break;
    case 'D':
      command = 'd';
      readonly = 1;
      break;
    case 't':
      token = optarg;
      break;
    case 'w':
      with = 1;
      break;
    case 'V':
      command = 'v';
      break;
    case 'h':
      command = 'h';
      break;
    case 'p':
      encrypted = 1;
      phrase = "";
      break;
    case 'P':
      encrypted = 1;
      phrase = optarg;
      break;
    case '?':
      return -1;
    }
  }

  /* check the sence of some options */
  if(token != "" && command != 'i') {
    cerr << pkg << ": -t only suitable with -i!\n";
    exit(1);
  }
  if(command == 'i' && cmd_insert != 1) {
    command = 'I'; /* read from stdin! */
  }
  else if(command == 'X') {
    command = 'h'; /* no command mode specified, so print out the usage message and exit */
  }
  if(with == 1 && command != 'S' && command != 's') {
    cerr << pkg << ": -w only suitable with -S or -s!\n";
    exit(1);
  }


  return 0;
}

